// file_io
// cs161 assignment 2 (fall 2005)
//
// kiran
//
// implementation (see header)
//

#include "file_io.h"
#include "logind.h"
#include "session.h"


extern logind server;



// fio_handler
// -------------------

// constructor
//
fio_handler::fio_handler(io_event_loop& loop, int in_fd, int out_fd)
  : io_handler(loop)
{
    fdin = in_fd;
    fdout = out_fd;

    iohandle *iohin = new iohandle(fdin);
    set_ioh(*iohin);
    iohin->set_blocking(false);
    want_read(true);

    iohandle *iohout = new iohandle(fdout);
    iohout->set_blocking(false);
}



// destructor
//
fio_handler::~fio_handler()
{
    close();
}



// ravail()
//
// called by the event loop when there is data on the fd
//
void fio_handler::ravail()
{
    int         i;
    fio_rsp     *rsp;

    // try to read a response on the pipe
    i = read(fdin, &rsp, sizeof(fio_rsp *));
    if (i != sizeof (fio_rsp *))
    {
        cerr << endl << "fio_handler incomplete response" << endl;
        return;
    }

    // this is risky: what if the session was deleted?
    rsp->requester->fio_response(rsp);
}



// request()
//
// called by a session to request a file i/o transaction
//
bool fio_handler::request(fio_req *req)
{
    int i = write(fdout, &req, sizeof(fio_req *));
    return (i == sizeof(fio_req *));
}



// fio_thread
// -------------------

// constructor
//
fio_thread::fio_thread(int in_fd, int out_fd)
{
    fdin = in_fd;
    fdout = out_fd;
}		



// destructor
//
fio_thread::~fio_thread()
{
}



// run()
//
// thread function
//
void fio_thread::run(void)
{
    int         i;
    fio_req     *req;
    fio_rsp     *rsp;

    for (; ;)
    {
        // get the next request
        i = read(fdin, &req, sizeof(fio_req *));
        if (i != sizeof(fio_req *))
            break;

        // process it
        rsp = new fio_rsp;
        switch (req->type)
        {
          case fio_req::LOGIN:
            rsp->ok = check_login(req->username, req->password);
            break;

          case fio_req::LIST:
            rsp->ok = do_list(req->username, req->largs.refname,
                                req->largs.mbname, &rsp->resp);
            break;

          default:
            break;
        }

        // done with the request, send the response
        rsp->req_type = req->type;
        rsp->requester = req->requester;
        rsp->tag = req->tag;
        delete req;
        i = write(fdout, &rsp, sizeof(fio_rsp *));
        if (i != sizeof(fio_rsp *))
            break;
    }

    cerr << endl << "fio_thread terminating" << endl;
}



// check_login()
//
// looks for the username in the password file and confirms
// that the correct password has been specified
//
bool fio_thread::check_login(string& username, string& password)
{
    bool        ok;
    string      user, pass;
    ifstream    ifpwd;


    // open the password file
    ifpwd.open(server.passwd_file.c_str());
    if (!ifpwd.good())
    {
        cerr << endl << "fio_thread: could not open password file" << endl;
        return false;
    }


    // read loop
    ok = false;
    while (ifpwd.good())
    {
        getline(ifpwd, user, ':');
        getline(ifpwd, pass);

        // found the user?
        if (username == user)
        {
            // check password
            if (pass == password)
                ok = true;
            break;
        }
    }


    // done
    ifpwd.close();


    return ok;
}



// do_list()
//
// processes the LIST command for the user
//
bool fio_thread::do_list(string& username, string& refname, string& mbname,
                            vector<string *> *resp)
{
    string              *s, mdir;
    struct stat         fst;


    // remove quotes
    strip_quotes(refname);
    strip_quotes(mbname);


    // hieararchy delimiter request?
    if (mbname == "")
    {
        s = new string("(\\Noselect) \"/\" \"\"");
        resp->push_back(s);
        return true;
    }


    // safety checks
    if (refname.find("..") != string::npos
            ||  mbname.find("..") != string::npos
            ||  (refname.size() > 0  &&  refname[0] == '/'))
    {
        s = new string("LIST: illegal parameters");
        resp->push_back(s);
        return false;
    }


    // the user's maildir + the refname
    mdir = server.mail_dir + "/" + username + "/" + refname;


    // if the maildir[/refname] doesn't exist, or if it isn't
    // a directory, there's nothing to return
    if (stat(mdir.c_str(), &fst) != 0  ||  !S_ISDIR(fst.st_mode))
        return true;


    // append the '/' if needed
    if (mdir[mdir.size() - 1] != '/')
    {
        mdir += "/";
        refname += "/";
    }


    // process wild cards allowed in the mailbox name
    // TODO: use glob() or something here
    int ppos = mbname.find_first_of('%');
    int spos = mbname.find_first_of('*');


    // no wildcard gets at most one response
    if (ppos == string::npos  &&  spos == string::npos)
    {
        mdir += mbname;
        if (stat(mdir.c_str(), &fst) == 0)
        {
            if (S_ISDIR(fst.st_mode))
                s = new string("(\\Noselect) \"/\" ");
            else
                s = new string("(\\Noinferiors) \"/\" ");
            *s += refname + mbname;
            resp->push_back(s);
        }

        return true;
    }


    // we only support non-recursive trailing wildcards at the moment
    int n = mbname.size();
    if (mbname == "*"  ||  mbname == "%"
            ||  (n >= 2  &&  mbname[n - 2] == '/'
                &&  (mbname[n - 1] == '*'  ||  mbname[n - 1] == '%')))
    {
        mdir += mbname;
        mdir.erase(mdir.size() - 1);

        // adjust root name
        if (n >= 2)
        {
            int n = mbname.find_last_of('/');
            refname += mbname.substr(0, n + 1);
        }

        get_dir(mdir, refname, resp);
    }


    return true;
}



// strip_quotes()
//
// removes double quotes from the beginning and end of a LIST
// parameter
//
void fio_thread::strip_quotes(string& s)
{
    if (s.compare("\"\"") == 0)
        s = "";
    else if (s.size() > 0)
    {
        if (s[0] == '\"')
            s.erase(0, 1);

        int n = s.size();
        if (n > 0  &&  s[n - 1] == '\"')
            s.erase(n - 1);
    }
}



// get_dir()
//
// loads the contents of a directory into the vector, formatted
// correctly as a LIST response
//
bool fio_thread::get_dir(string& mdir, string& rootname,
                            vector<string *> *resp)
{
    string              *s, full;
    struct dirent       *dent;
    struct stat         fst;

    // open the directory
    DIR *md = opendir(mdir.c_str());
    if (md)
    {
        // read in entries in the directory
        while (dent = readdir(md))
        {
            full = mdir + dent->d_name;
            if (stat(full.c_str(), &fst) != 0)
                continue;

            if (S_ISDIR(fst.st_mode))
            {
                // skip . and ..
                if (strcmp(dent->d_name, ".") == 0
                        ||  strcmp(dent->d_name, "..") == 0)
                    continue;

                s = new string("(\\Noselect) \"/\" ");
            }
            else
                s = new string("(\\Noinferiors) \"/\" ");

            *s += rootname;
            *s += dent->d_name;
            if (S_ISDIR(fst.st_mode))
                *s += "/";
            resp->push_back(s);
        }

        // done
        closedir(md);
    }


    return true;
}
