// file_io
// cs161 assignment 2 (fall 2005)
//
// kiran
//
// helpers to perform file i/o (blocking) on behalf of sessions
//

#ifndef __FILE_IO_H__ 
#define __FILE_IO_H__


#include <dirent.h>

#include <nmstl/netioevent> 
#include <nmstl/terminal> 
#include <nmstl/ntime>

using namespace nmstl;


#include <iostream>
#include <fstream>

using namespace std;


// forward declarations
class session;


// arguments for LIST
struct list_args
{
    string              refname;
    string              mbname;
};


// fio_req: request
struct fio_req
{
    static const int    LOGIN = 0,
                        LIST = 1;

    int                 type;
    string              username;
    session             *requester;
    string              tag;

    string              password;
    list_args           largs;
};



// fio_rsp: response
struct fio_rsp
{
    int                 req_type;
    bool                ok;
    session             *requester;
    string              tag;

    vector<string *>    resp;
};



// fio_handler
// nmstl io_handler that talks to the file_io_thread
//
class fio_handler : public io_handler
{
  public:
    fio_handler(io_event_loop& loop, int, int);
    ~fio_handler();

    // send a request
    bool request(fio_req *);

    // callbacks
    void ravail(void);

  private:
    // request/response pipes
    int         fdin, fdout;
};



// fio_thread
// used with nmstl's threaded class
//
class fio_thread
{
  public:
    fio_thread(int, int);
    ~fio_thread();

    // thread function
    void run(void);

  private:
    // helpers
    bool check_login(string&, string&);
    bool do_list(string&, string&, string&, vector<string *> *);
    void strip_quotes(string&);
    bool get_dir(string&, string&, vector<string *> *);

    // request/response pipes
    int         fdin, fdout;
};


#endif  // __FILE_IO_H__
