// logind
// cs161 assignment 2 (fall 2005)
//
// kiran
//
// initialization code
//

#include "logind.h"
#include "session.h"
#include "file_io.h"

#include <unistd.h>

// single instance of logind
logind server;



// entry point
//
int main(int argc, char **argv)
{
    // start the server
    server.run(argc, argv);

    return 0;
}



// run()
//
// start the server
//
void logind::run(int argc, char **argv)
{
    // initialize instance members
    port = 0;
    passwd_file = "";
    mail_dir = "";
    max_fds = 0;
    num_fiohs = 0;
    cur_fds = 0;
    cur_fioh = 0;


    // parse command line arguments
    if (argc != 9)
        usage();
    for (int i = 2;  i < argc;  i += 2)
    {
        if (strcasecmp(argv[i-1], "-p") == 0)
            port = atoi(argv[i]);
        else if (strcasecmp(argv[i-1], "-f") == 0)
            passwd_file = argv[i];
        else if (strcasecmp(argv[i-1], "-m") == 0)
            mail_dir = argv[i];
        else if (strcasecmp(argv[i-1], "-n") == 0)
            max_fds = atoi(argv[i]);
        else
            usage();
    }
    if (port < 1024  ||  max_fds == 0
            ||  passwd_file.compare("") == 0  ||  mail_dir.compare("") == 0)
        usage();


    // create event loop
    loop = new io_event_loop();


    // create terminal
    term = new term_handler(*loop, "logind> ");
    *term << "starting login server on port " << port << "..." << endl;


    // create file i/o helpers
    int                         h_to_t[2], t_to_h[2];
    threaded<fio_thread>        *fthr;
    fio_handler                 *fioh;

    num_fiohs = max(1, (max_fds / FDS_PER_THREAD));
    fiohs = new vector<fio_handler *>(num_fiohs);
    fiohs->clear();
    for (int i = 0;  i < num_fiohs;  i++)
    {
        // pipes for communication between handler and thread
        if (pipe(h_to_t) != 0  ||  pipe(t_to_h) != 0)
        {
            cerr << endl << "pipe() failed, terminating" << endl;
            exit(-1);
        }

        // handler goes in the vector for later use
        fioh = new fio_handler(*loop, t_to_h[0], h_to_t[1]);
        fiohs->push_back(fioh);

        // thread is started immediately
        fthr = new threaded<fio_thread>(h_to_t[0], t_to_h[1]);
        fthr->start();
    }


    // start the TCP acceptor
    tcp_acceptor<session, term_handler *> spawner(*loop, port, term);


    // start the event loop
    loop->run();
}



// get_fioh()
//
// called by a session to request a fio_handler to use
// for a received file i/o request
//
fio_handler *logind::get_fioh(void)
{
    cur_fioh = (cur_fioh + 1) % num_fiohs;
    return fiohs->at(cur_fioh);
}



// usage()
//
// displays usage message and exits
//
void logind::usage(void)
{
    cout << "Usage:" << endl
         << "logind -p <port> -f <passwd_file -m <mail_dir> -n <max_fds>"
         << endl;
    exit(-1);
}


